# coding: utf-8
"""
    dooku.tests
    ~~~~~~~~~~~

    Tests Dooku itself.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""
import unittest

try:
    from unittest import mock  # NOQA
except:
    import mock  # NOQA


class DookuTestCase(unittest.TestCase):
    """
    The base class for all Dooku's test cases.
    """



# -*- coding: utf-8 -*-
"""
    Sphinx configuration file for building dooku's documentation.
"""
from __future__ import unicode_literals

import re
import os
import sys


# add parent dir to PYTHONPATH for allowing import em's version
sys.path.append(os.path.abspath(os.pardir))
from dooku import __version__ as dooku_version


# project settings
project = 'dooku'
copyright = '2014, Igor Kalnitsky'
release = dooku_version
version = re.sub('[^0-9.]', '', release)

# sphinx settings
extensions = ['sphinx.ext.autodoc']
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['_build']
pygments_style = 'sphinx'

# html output settings
html_theme = 'default'
html_static_path = ['_static']



# coding: utf-8
"""
    dooku.tests.test_datetime
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests Dooku's datetime library.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""
import re
import time
import datetime

from dooku.datetime import UTC, Local, to_iso8601
from . import DookuTestCase


class TestDatetimeTz(DookuTestCase):
    def test_utc_is_instance(self):
        """
        The UTC tz imlementation has to be an instance of built-in
        :class:`~datetime.tzinfo` class.
        """
        self.assertIsInstance(UTC, datetime.tzinfo)

    def test_local_is_instance(self):
        """
        The Local tz imlementation has to be an instance of built-in
        :class:`~datetime.tzinfo` class.
        """
        self.assertIsInstance(Local, datetime.tzinfo)

    def test_utc_timezone(self):
        """
        The UTC tz implementation has to meet next criteria:

        * utcoffset is 0 (there's no time shift)
        * dst is 0 (off)
        * datetime's timetuple and utctimetuple are equal
        """
        now = datetime.datetime.now(UTC)

        self.assertEqual(now.utcoffset(), datetime.timedelta(0))
        self.assertEqual(now.dst(), datetime.timedelta(0))
        self.assertEqual(now.utctimetuple(), now.timetuple())

    def test_local_timezone(self):
        """
        The Local tz implementation has to meet next criteria:

        * utcoffset has to be either time.timezone or time.altzone depends
          on time.daylight value;
        * dst has to be either 0 or difference between time.altzone and
          time.timezone;
        * datetime's timetuple and utctimetuple are not equal
        """
        l_dst = datetime.timedelta(0)
        l_timezone = datetime.timedelta(seconds=-time.timezone)

        if time.daylight:
            l_dst = datetime.timedelta(seconds=time.altzone - time.timezone)
            l_timezone = datetime.timedelta(seconds=-time.altzone)

        now = datetime.datetime.now(Local)
        self.assertEqual(now.utcoffset(), l_timezone)
        self.assertEqual(now.dst(), l_dst)

    def test_local_utc_convertion(self):
        """
        UTC <-> Local timzone convertions should be consistent.
        """
        now = datetime.datetime.now(UTC)
        lcl = datetime.datetime.now(Local)

        self.assertEqual(
            now.astimezone(Local).astimezone(UTC), now)

        self.assertEqual(
            lcl.astimezone(UTC).astimezone(Local), lcl)


class TestDatetimeFormat(DookuTestCase):
    re_iso8601 = re.compile(
        r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{1,})?(Z|[-+]\d{2}:\d{2})$'
    )

    def test_re_iso8601(self):
        """
        The re_iso8601 regex has to work correct and without mistakes, since
        it's a judge of other test cases.
        """
        dataset = (
            # valid set
            ('2014-05-24T17:16:16Z', self.assertRegex),
            ('2014-05-24T17:16:16.42Z', self.assertRegex),
            ('2014-05-24T17:16:16.42+01:00', self.assertRegex),
            ('2014-05-24T17:16:16+01:00', self.assertRegex),
            ('2014-05-24T17:16:16-01:00', self.assertRegex),
            ('2014-05-24T17:16:16.42+01:00', self.assertRegex),
            ('2014-05-24T17:16:16.42-01:00', self.assertRegex),

            # invalid set
            ('2014-05-24T17:16', self.assertNotRegex),
            ('2014-05-24T17:16:16', self.assertNotRegex),
            ('2014-05-24T17:16:16.Z', self.assertNotRegex),
            ('2014-05-24T17:16:16.42', self.assertNotRegex),
            ('2014-05-24T17:16:16+01:001', self.assertNotRegex),
            ('2014-05-24T17:16:16+001:00', self.assertNotRegex),
            ('2014-05-24T17:16:16+01:1', self.assertNotRegex),
            ('2014-05-24T17:16:16+1:01', self.assertNotRegex),
            ('2014-05-24T17:16:16+1:1', self.assertNotRegex),
            ('2014-05-24T17:16:16+01:20Z', self.assertNotRegex),
        )

        for text, assert_fn in dataset:
            assert_fn(text, self.re_iso8601)

    def test_iso8601(self):
        """
        The to_iso8601 has to produce a valid output for naive datetime
        objects.
        """
        now = datetime.datetime.now()
        now = to_iso8601(now)

        self.assertRegex(now, self.re_iso8601)

    def test_iso8601_utc(self):
        """
        The to_iso8601 has to produce a valid output for aware datetime
        objects.
        """
        now = datetime.datetime.now(UTC)
        now = to_iso8601(now)

        self.assertRegex(now, self.re_iso8601)
        self.assertTrue(now.endswith('Z') or now.endswith('+00:00'))

    def test_iso8601_local(self):
        """
        The to_iso8601 has to produce a valid output for aware datetime
        objects.
        """
        now = datetime.datetime.now(Local)
        now = to_iso8601(now)

        self.assertRegex(now, self.re_iso8601)
        self.assertTrue(not now.endswith('Z'))



# coding: utf-8
"""
    dooku.conf
    ~~~~~~~~~~

    The module provides a useful class to deal with configuration data.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""
import copy
import itertools


class Conf(object):
    """
    A :class:`dict` wrapper that extends its functionality.

    The class was designed with one purpose: to provide the more convenient
    way to interact with dictionary instances that keep some configuration
    data. In many ways it behaves exactly like a dictionary, but unlike the
    last one it can be accessed by a *compound key* and supports *recursive
    update*.

    A Compound Key is a key that consists of two or more keys that allow
    you to retrieve value from subnode of subnode. For example, we have
    the following structure ::

        { 'ukraine': {
            'kharkiv': {
              'timezone': 'UTC+2', }, }, }

    and you want to get the timezone's value with one query. Well, with
    compound keys you can do it this way ::

        timezone = conf['ukraine.kharkiv.timezone']

    instead of usual ::

        timezone = conf['ukraine']['kharkiv']['timezone']

    On the other hand we have a usual dictionary behaviour:

    * a ``KeyError`` exception will be thrown in case of key absence
    * a :meth:`get` method is provided to retrieve values safely

    The class overrides an :meth:`update` method so it can be used to
    recursively update values. For example, we have the following structure ::

        { 'ukraine': {
            'kharkiv': {
              'timezone': 'UTC+2', }, }, }

    and want to update it with next one ::

        { 'ukraine': {
            'kharkiv': {
              'population': '1 427 000', }, }, }

    In vanilla Python's dictionary we will have the same result as statement
    above since it completely overwrites top-level keys, but our class can
    handle it properly and merge it automatically, so we will have next
    result ::

        { 'ukraine': {
            'kharkiv': {
              'timezone': 'UTC+2',
              'population': '1 427 000', }, }, }

    :param confs:
        A list of dictionaries to create a Conf instance based on it.
        Each next dictionary overrides settings from the previous one.
    :param separator:
        A character that's used as separator in compound keys
    """

    #: this character will be used as a separator in case you don't
    #: set a new one explicitly
    default_separator = '.'

    def __init__(self, *confs, **options):
        self._data = {}
        self._separator = options.get('separator', self.default_separator)

        for conf in confs:
            self.update(copy.deepcopy(conf))

    def update(self, iterable={}, **kwargs):
        """
        Updates recursively a self with a given iterable.

        TODO: rewrite this ugly stuff
        """
        def _merge(a, *args):
            for key, value in itertools.chain(*args):
                if key in a and isinstance(value, (dict, Conf)):
                    value = _merge(a[key], value.items())
                a[key] = value
            return a

        # adopt iterable sequence to unified interface: (key, value)
        if isinstance(iterable, (dict, Conf)):
            iterable = iterable.items()

        # iterate and update values
        _merge(self._data, iterable, kwargs.items())

    def items(self):
        """
        Returns an iterator of a sequence of pairs ``(key, value)``, just
        like it works for ``dict``.

        :returns: (iter) an iterator of a sequence of key-value pairs
        """
        return self._data.items()

    def get(self, compound_key, default=None):
        """
        Returns a value that's associated with a given compound key.

        Unlike :meth:`__getitem__`, the method provides fallback to default
        in case key doesn't exist.

        :param compound_key: (str) a key for retrieving value
        :param default: (object) a default value
        :returns: (object) retrieved value if key exists; otherwise - default
        """
        try:
            value = self[compound_key]
        except KeyError:
            value = default
        return value

    def __getitem__(self, compound_key):
        """
        Returns a value that's associated with a given compound key.

        :param compound_key: (str) a key for retrieving value
        :returns: (object) retrieved value if key exists
        :raises KeyError: a given key does not exist
        """
        conf = self._data

        for key in compound_key.split(self._separator):
            conf = conf[key]

        # We need to return dict as Conf instance to make possible use Conf's
        # features on it. The returned object is used exactly like a wrapper,
        # since it doesn't make a full copy of a given dict.
        if isinstance(conf, dict):
            rv = Conf(separator=self._separator)
            rv._data = conf
            return rv

        return conf

    def __setitem__(self, compound_key, value):
        """
        Sets a value for given option.

        :param compound_key: (str) an option to change
        :param value: (object) a value to set
        :raises KeyError: an option doesn't exist
        """
        conf = self._data
        keys = compound_key.split(self._separator)
        for key in keys[:-1]:
            if key not in conf:
                conf[key] = {}
            conf = conf[key]
        conf[keys[-1]] = value

    def __delitem__(self, compound_key):
        """
        Remove a given compound key from the instance.

        :param compound_key: (str) a key to delete
        """
        conf = self._data
        keys = compound_key.split(self._separator)
        for key in keys[:-1]:
            conf = conf[key]
        del conf[keys[-1]]

    def __contains__(self, compound_key):
        """
        Checks if a given compound key exists.

        :param compound_key: (str) a key to check
        :returns: (bool) True if exists; otherwise - False
        """
        try:
            self[compound_key]
        except KeyError:
            return False
        return True

    def __eq__(self, other):
        """
        Equality operator implementation makes us possible to compare
        Conf instances with other Conf instances, or even dictionaries.

        :param other: (dict, Conf) an instance to compare with
        :returns: True if `self` is equal to `other`; otherwise - False
        """
        if isinstance(other, dict):
            return self._data == other
        elif isinstance(other, Conf):
            return self._data == other._data
        return False

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return '<{0} {1}>'.format(self.__class__.__name__, repr(self._data))



# coding: utf-8
"""
    dooku.datetime
    ~~~~~~~~~~~~~~

    The package provides some helper things that you might need for datetime
    manipulations: formatters and tz classes.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""

# despite the fact we use python3 only, we still need this statement
# in order to do not fail during sphinx execution.
from __future__ import absolute_import


import re
import time
import datetime


def to_iso8601(dt, tz=None):
    """
    Returns an ISO-8601 representation of a given datetime instance.

        >>> to_iso8601(datetime.datetime.now())
        '2014-10-01T23:21:33.718508Z'

    :param dt: a :class:`~datetime.datetime` instance
    :param tz: a :class:`~datetime.tzinfo` to use; if None - use a default one
    """
    if tz is not None:
        dt = dt.replace(tzinfo=tz)
    iso8601 = dt.isoformat()

    # Naive datetime objects usually don't have info about timezone.
    # Let's assume it's UTC and add Z to the end.
    if re.match(r'.*(Z|[+-]\d{2}:\d{2})$', iso8601) is None:
        iso8601 += 'Z'

    return iso8601

#: The RFC-3339 is a profile (subset) of more complex ISO-8601. So it's
#: just an alias for :func:`to_iso8601` function.
to_rfc3339 = to_iso8601


class UTC(datetime.tzinfo):
    """
    Implements a UTC :class:`datetime.tzinfo`.

    You can use it to attach timezone information to datetime instances.
    Such information becomes useful if you want to represent a datetime
    instance in various timezone notations.

    Example:

        >>> from dooku.datetime import UTC
        >>> dt = datetime.datetime.now(UTC)
        >>> dt
        datetime.datetime(2014, 5, 29, 20, 22, 17, 426248, tzinfo=<...>)

    """
    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def dst(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return 'UTC'
UTC = UTC()


class Local(datetime.tzinfo):
    """
    Implements a Local :class:`datetime.tzinfo`.

    You can use it to attach timezone information to datetime instances.
    Such information becomes useful if you want to represent a datetime
    instance in various timezone notations.

    Example:

        >>> from dooku.datetime import UTC, Local
        >>> dt = datetime.datetime.now(UTC)
        >>> dt
        datetime.datetime(2014, 5, 29, 20, 22, 17, 426248, tzinfo=<...>)
        >>>
        >>> dt = dt.astimezone(Local)
        >>> dt
        datetime.datetime(2014, 5, 29, 23, 22, 17, 426248, tzinfo=<...>)
    """
    @staticmethod
    def _is_dst(dt):
        """
        Returns True if a given datetime object represents a time with
        DST shift.
        """
        # we can't use `dt.timestamp()` here since it requires a `utcoffset`
        # and we don't want to get into a recursive loop
        localtime = time.localtime(time.mktime((
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            dt.weekday(),
            0,              # day of the year
            -1              # dst
        )))
        return localtime.tm_isdst > 0

    def utcoffset(self, dt):
        if self._is_dst(dt):
            return datetime.timedelta(seconds=-time.altzone)
        return datetime.timedelta(seconds=-time.timezone)

    def dst(self, dt):
        """
        Returns a difference in seconds between standard offset and
        dst offset.
        """
        if not self._is_dst(dt):
            return datetime.timedelta(0)

        offset = time.timezone - time.altzone
        return datetime.timedelta(seconds=-offset)

    def tzname(self, dt):
        return 'Local Time'
Local = Local()



# coding: utf-8
"""
    dooku
    ~~~~~

    Dooku is a set of libraries for the Python programming language that
    provides various useful stuff that you might need in everyday usage.

    The idea is to provide a set of libraries and useful functions, just
    like Boost does in C++ world.

    Dooku is a good not only as a set of libraries, it also may be used
    as a some kind of incubator where various developers can test and
    evolve their projects. So please don't hesitate to propose your
    library to be a part of Dooku.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""

__version__ = '0.1.0'
__license__ = 'BSD'



# coding: utf-8
"""
    dooku.algorithm
    ~~~~~~~~~~~~~~~

    The module implements various algorithms for different purposes.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""


def any_of(pred, iterable):
    """
    Returns ``True`` if ``pred`` returns ``True`` for any of the elements in
    the ``iterable`` range, and ``False`` otherwise.

        >>> any_of(lambda x: x == 4, [1, 2, 3, 4])
        True

    :param pred: a predicate function to check a value from the iterable range
    :param iterable: an iterable range to check in
    :returns: ``True`` if ``pred`` returns ``True`` for any of the elements
    """
    return any((pred(i) for i in iterable))


def all_of(pred, iterable):
    """
    Returns ``True`` if ``pred`` returns ``True`` for all the elements in
    the ``iterable`` range or if the range is empty, and ``False`` otherwise.

        >>> all_of(lambda x: x % 2 == 0, [2, 4, 6, 8])
        True

    :param pred: a predicate function to check a value from the iterable range
    :param iterable: an iterable range to check in
    :returns: ``True`` if ``pred`` returns ``True`` for all the elements
    """
    return all((pred(i) for i in iterable))


def find_if(pred, iterable, default=None):
    """
    Returns a reference to the first element in the ``iterable`` range for
    which ``pred`` returns ``True``. If no such element is found, the
    function returns ``default``.

        >>> find_if(lambda x: x == 3, [1, 2, 3, 4])
        3

    :param pred: a predicate function to check a value from the iterable range
    :param iterable: an iterable range to check in
    :param default: a value that will be returned if no elements were found
    :returns: a reference to the first found element or default
    """
    return next((i for i in iterable if pred(i)), default)


def none_of(pred, iterable):
    """
    Returns ``True`` if ``pred`` returns ``False`` for all the elements in
    the ``iterable`` range or if the range is empty, and ``False`` otherwise.

        >>> none_of(lambda x: x % 2 == 0, [1, 3, 5, 7])
        True

    :param pred: a predicate function to check a value from the iterable range
    :param iterable: an iterable range to check in
    :returns: ``True`` if all elements don't satfisfy ``pred``
    """
    return all((not pred(i) for i in iterable))



# coding: utf-8
"""
    dooku.ext
    ~~~~~~~~~

    The module provides things that are allow you to load dynamically
    extensions. The idea is based on Python's ``entry_points``.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""
import pkg_resources


class ExtensionManager(object):
    """
    Load and manage your extensions with fun!

    The class was designed to provide minimal functionality to interact
    with extension collections. In many ways it works similar to
    stevedore_, but unlike the last one it operates with *exported beings*
    and doesn't create instances for those beings.

    Well, what it means for you? It means...

    * You can discover and load extensions from a given namespace.
    * You can get extension (exported being) by name.
    * You can iterate over discovered and loaded extensions.
    * You can check whether extension is loaded or not.

    Behind the class lies an idea to discover extensions by means of
    `entry_points`_. So the first thing you have to do is to declare
    your plugin in ``setup.py`` and then install it::

        sutup(
            # ...
            entry_points={
                'my_plugin_namespace': [
                    'plugin_name = my_package.my_plugin:MyPluginClass',
                ]
            }
        )

    After plugin installation it becomes available for discovering and
    loading, so you probably want to get it::

        from dooku.ext import ExtensionManager

        for name, extension in ExtensionManager('my_plugin_namespace'):
            # name is plugin_name
            # extension is MyPluginClass

    You also can load plugins selectively. Look at class parameters for
    details.

    :param namespace:
        A namespace to import from as string.
    :param names:
        A list of objects to import; imports all if ``None``.
    :param silent:
        Skip loading errors if ``True``; otherwise - throw exception.

    .. _stevedore:    https://stevedore.readthedocs.org/
    .. _entry_points: https://pythonhosted.org/setuptools/setuptools.html
                      #dynamic-discovery-of-services-and-plugins
    """
    def __init__(self, namespace, names=None, silent=False):
        #: `name` <-> `extensions list` map
        #:
        #: Since extension is an exported object and know nothing about
        #: it's name, we have to save this info here for further usage.
        self._extensions = {}

        # search for extensions in a given namespace
        for entry_point in pkg_resources.iter_entry_points(namespace):
            if names is not None and entry_point.name not in names:
                continue

            try:
                ext = entry_point.load()

                if entry_point.name not in self._extensions:
                    self._extensions[entry_point.name] = []
                self._extensions[entry_point.name].append(ext)

            except Exception:
                if not silent:
                    raise

    def get(self, name, default=None):
        """
        Returns an extension instance with a given name.

        In case there are few extensions with a given name, the first one
        will be returned. If no extensions with a given name are exist,
        the `default` value will be returned.

        :param name: (str) an extension name
        :param default: (object) a fallback value
        :returns: (object) an extension instance
        """
        try:
            value = self[name]
        except KeyError:
            value = default
        return value

    def getall(self, name):
        """
        Returns an extension instances with a given name.

        If no extensions with a given name are exist, an empty list will
        be returned.

        :param name: (str) an extension name
        :returns: (list of object) a list of extension instances
        """
        # we're interested to return a copy to protect us
        # from unexpected modifications
        return list(self._extensions.get(name, []))

    def names(self):
        """
        Returns a list of plugin names that were loaded.

        :returns: (set) plugin names
        """
        # convert dict_keys type to set
        return set(self._extensions.keys())

    def __getitem__(self, name):
        """
        Returns an extension instance with a given name.

        In case there are few extensions with a given name, the first one
        will be returned.

        :param name: (str) an extension name
        :returns: (object) an extension instance
        """
        # we always have at least one item in the list
        return self._extensions[name][0]

    def __contains__(self, name):
        """
        Tests whether exist an extension with a given name or not.

        :param name: (str) an extension name
        :returns: (bool) True if exist; otherwise - False
        """
        return name in self._extensions

    def __iter__(self):
        """
        Returns an iterator to extensions collection.

        :returns: (object) an iterator over extensions
        """
        for key in self._extensions:
            for value in self._extensions[key]:
                yield key, value



# coding: utf-8
"""
    dooku.tests.test_decorator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests Dooku's decorator stuff.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""

from dooku.decorator import cached_property

from . import DookuTestCase


class TestCachedProperty(DookuTestCase):

    def setUp(self):
        """
        Creates a test class with a cached property.
        """
        class Foo(object):
            @cached_property
            def bar(self):
                return object()
        self.Foo = Foo

    def test_it_wraps(self):
        """
        Test that a method was converted to a cached_property.
        """
        self.assertIsInstance(self.Foo.bar, cached_property)

    def test_it_returns_the_same_instance(self):
        """
        Test that a cached property always returns the same instance.
        """
        foo = self.Foo()

        instance_a = foo.bar
        instance_b = foo.bar

        self.assertEqual(instance_a, instance_b)
        self.assertIs(instance_a, instance_b)

    def test_various_objects_have_various_instances(self):
        """
        Test that a cached property of various objects returns various
        instances.
        """
        foo_a = self.Foo()
        foo_b = self.Foo()

        instance_a = foo_a.bar
        instance_b = foo_b.bar

        self.assertNotEqual(instance_a, instance_b)
        self.assertIsNot(instance_a, instance_b)



#!/usr/bin/env python
# coding: utf-8
"""
Dooku
-----

Dooku is a set of libraries for the Python programming language that
provides various useful stuff that you might need in everyday usage.

The idea is to provide a set of libraries and useful functions, just
like Boost does in C++ world.

Dooku is a good not only as a set of libraries, it also may be used
as a some kind of incubator where various developers can test and
evolve their projects. So please don't hesitate to propose your
library to be a part of Dooku.

Links
`````

* `documentation <https://dooku.readthedocs.org/>`_
* `source code <https://github.com/ikalnitsky/dooku>`_
"""
from setuptools import setup, find_packages

from dooku import __version__ as dooku_version
from dooku import __license__ as dooku_license


setup(
    name='dooku',
    version=dooku_version,
    license=dooku_license,
    description='a set of libraries for everyday',
    long_description=__doc__,
    url='https://github.com/ikalnitsky/dooku',
    platforms=['any'],

    author='Igor Kalnitsky',
    author_email='igor@kalnitsky.org',

    packages=find_packages(exclude=('tests')),
    test_suite='tests',

    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',

        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',

        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ],
)



# coding: utf-8
"""
    dooku.tests.test_algorithm
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests Dooku's algorithm stuff.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""

from dooku.algorithm import *  # NOQA

from . import DookuTestCase


class TestStdCpp(DookuTestCase):

    def test_any_of(self):
        """
        The any_of function has to return expected result for all samples
        of test set.
        """
        self.assertTrue(any_of(lambda x: x == 1, [1, 2, 3, 4, 5]))
        self.assertTrue(any_of(lambda x: x == 2, [1, 2, 3, 4, 5]))
        self.assertTrue(any_of(lambda x: x == 5, [1, 2, 3, 4, 5]))

        self.assertFalse(any_of(lambda x: x == 0, [1, 2, 3, 4, 5]))
        self.assertFalse(any_of(lambda x: x == 6, [1, 2, 3, 4, 5]))

    def test_all_of(self):
        """
        The all_of function has to return expected result for all samples
        of test set.
        """
        self.assertTrue(all_of(lambda x: x > 0, [1, 2, 3, 4, 5]))
        self.assertTrue(all_of(lambda x: x < 6, [1, 2, 3, 4, 5]))

        self.assertFalse(all_of(lambda x: x > 1, [1, 2, 3, 4, 5]))
        self.assertFalse(all_of(lambda x: x == 2, [1, 2, 3, 4, 5]))

    def test_none_of(self):
        """
        The none_of function has to return expected result for all samples
        of test set.
        """
        self.assertTrue(none_of(lambda x: x == 7, [1, 2, 3, 4, 5]))

        self.assertFalse(none_of(lambda x: x == 1, [1, 2, 3, 4, 5]))
        self.assertFalse(none_of(lambda x: x == 2, [1, 2, 3, 4, 5]))
        self.assertFalse(none_of(lambda x: x == 5, [1, 2, 3, 4, 5]))

    def test_find_if(self):
        """
        The find_if function has to return expected result for all samples
        of test set.
        """
        self.assertEqual(find_if(lambda x: x == 1, [1, 2, 3, 4, 5]), 1)
        self.assertEqual(find_if(lambda x: x == 2, [1, 2, 3, 4, 5]), 2)
        self.assertEqual(find_if(lambda x: x == 5, [1, 2, 3, 4, 5]), 5)
        self.assertIsNone(find_if(lambda x: x == 0, [1, 2, 3, 4, 5]))

        vector = [{'a': 13}, {'a': 42}]
        result = find_if(lambda x: x['a'] == 42, vector)

        self.assertIsInstance(result, dict)
        self.assertEqual(id(result), id(vector[1]))



# coding: utf-8
"""
    dooku.tests.test_ext
    ~~~~~~~~~~~~~~~~~~~~

    Tests Dooku's ext-related stuff.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""
from dooku.ext import ExtensionManager
from . import DookuTestCase, mock


class TestExtensionManager(DookuTestCase):

    @mock.patch('dooku.ext.pkg_resources.iter_entry_points')
    def setUp(self, iter_ep):
        """
        Prepare extensions for each testcase.
        """
        # Holy crap! The mock library can't work with "name" attribute,
        # so we need this tricks to workaround this.
        entry_points = [mock.Mock(), mock.Mock(), mock.Mock()]
        type(entry_points[0]).name = mock.PropertyMock(return_value='one')
        type(entry_points[1]).name = mock.PropertyMock(return_value='two')
        type(entry_points[2]).name = mock.PropertyMock(return_value='two')

        iter_ep.return_value = entry_points

        self.ext_manager = ExtensionManager('dooku.test')
        self.entry_points = entry_points

    def test_get(self):
        """
        The get method has to behave exactly like a dict's one.
        """
        self.assertEqual(
            self.ext_manager.get('one'), self.entry_points[0].load())
        self.assertEqual(
            self.ext_manager.get('two'), self.entry_points[1].load())
        self.assertNotEqual(
            self.ext_manager.get('two'), self.entry_points[2].load())

        self.assertIsNone(self.ext_manager.get('three'))
        self.assertEqual(self.ext_manager.get('three', 42), 42)

    def test_getall(self):
        """
        The getall method has to return a list of extensions for a given
        name. In case there're no extensions - the empty list has to be
        returned.
        """
        self.assertEqual(
            self.ext_manager.getall('one'), [self.entry_points[0].load()])
        self.assertEqual(
            self.ext_manager.getall('two'),
            [self.entry_points[1].load(), self.entry_points[2].load()])

        self.assertEqual(self.ext_manager.getall('three'), [])

    def test_getall_returns_copy(self):
        """
        The getall method has to return a list copy so its changes has
        no affect internal state.
        """
        res = self.ext_manager.getall('two')
        res.append(42)

        self.assertNotEqual(self.ext_manager.getall('two'), res)

    def test_getitem(self):
        """
        The __getitem__ has to behave exactly like a dict's one.
        """
        self.assertEqual(
            self.ext_manager['one'], self.entry_points[0].load())
        self.assertEqual(
            self.ext_manager['two'], self.entry_points[1].load())
        self.assertNotEqual(
            self.ext_manager['two'], self.entry_points[2].load())

        self.assertRaises(KeyError, lambda: self.ext_manager['three'])

    def test_names(self):
        """
        The names method has to return a set of extension names.
        """
        self.assertEqual(self.ext_manager.names(), set(['one', 'two']))

    def test_contains(self):
        """
        The __contains__ has to behave exactly like a dict's one.
        """
        self.assertIn('one', self.ext_manager)
        self.assertIn('two', self.ext_manager)
        self.assertNotIn('three', self.ext_manager)

    def test_iter(self):
        """
        The __iter__ has to return an iterator over all extensions where the
        iteration element is (name, extension).
        """
        self.assertEqual(set(i for i in self.ext_manager), set([
            ('one', self.entry_points[0].load()),
            ('two', self.entry_points[1].load()),
            ('two', self.entry_points[2].load()), ]))

    @mock.patch('dooku.ext.pkg_resources.iter_entry_points')
    def test_constructor_w_names(self, iter_ep):
        """
        If the names was passed to the constructor, only those extensions
        has to be loaded.
        """
        iter_ep.return_value = self.entry_points
        self.ext_manager = ExtensionManager('dooku.test', ['two'])

        self.assertEqual(set(i for i in self.ext_manager), set([
            ('two', self.entry_points[1].load()),
            ('two', self.entry_points[2].load()), ]))

    @mock.patch('dooku.ext.pkg_resources.iter_entry_points')
    def test_silent_false(self, iter_ep):
        """
        The constructor has to raise exceptions if silent is False.
        """
        self.entry_points[1].load.side_effect = ValueError('error')
        iter_ep.return_value = self.entry_points

        self.assertRaises(
            ValueError, lambda: ExtensionManager('dooku.test', silent=False))

    @mock.patch('dooku.ext.pkg_resources.iter_entry_points')
    def test_silent_true(self, iter_ep):
        """
        The constructor don't has to raise exceptions if silent is True.
        """
        self.entry_points[0].load.side_effect = ValueError('error')
        iter_ep.return_value = self.entry_points

        self.ext_manager = ExtensionManager('dooku.test', silent=True)

        self.assertEqual(set(i for i in self.ext_manager), set([
            ('two', self.entry_points[1].load()),
            ('two', self.entry_points[2].load()), ]))
        self.assertEqual(self.ext_manager.names(), set(['two']))



# coding: utf-8
"""
    dooku.tests.test_conf
    ~~~~~~~~~~~~~~~~~~~~~

    Tests Dooku's conf-related stuff.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""
from dooku.conf import Conf

from . import DookuTestCase


class TestConf(DookuTestCase):

    def setUp(self):
        """
        Prepare source config for each testcase.
        """
        self.source_conf = {
            'root': {
                'one': {
                    'a': 1,
                    'b': 2,
                },

                'two': {
                    'c': 3,
                }
            }
        }

    def test_constructor_saves_dict(self):
        """
        The constructor has to save an input dict without changes.
        """
        conf = Conf(self.source_conf)
        self.assertEqual(self.source_conf, conf._data)

    def test_constructor_makes_copy(self):
        """
        The constructor has to make a copy of input dictionaries.
        """
        conf = Conf(self.source_conf)
        conf['non-root'] = 13
        conf['root.one.c'] = 42

        self.assertNotIn('non-root', self.source_conf)
        self.assertNotIn('c', self.source_conf['root']['one'])

    def test_constructor_custom_separator(self):
        """
        The constructor has to use a given separator, not a dot.
        """
        conf = Conf(self.source_conf, separator='#')

        self.assertEqual(conf['root#one#a'], conf['root']['one']['a'])
        self.assertEqual(conf.get('root#one#a'), conf['root']['one']['a'])

        self.assertEqual(conf['root#two#c'], conf['root']['two']['c'])
        self.assertEqual(conf.get('root#two#c'), conf['root']['two']['c'])

    def test_change_subconf(self):
        """
        Tests that changing some option via sub-conf instance leads
        to change this option in the main conf instance.
        """
        conf = Conf(self.source_conf)

        one_conf = conf['root.one']
        one_conf['a'] = 42

        self.assertEqual(one_conf['a'], 42)
        self.assertEqual(conf['root.one.a'], 42)

    def test_update(self):
        """
        The update has to be capable to make recursive update.
        """
        conf = Conf(self.source_conf)

        conf.update({
            'root': {
                'one': {'a': 42, 'z': 13},
                'three': [],
            },
            'non-root': [1, 2],
        })

        result = {
            'root': {
                'one': {'a': 42, 'b': 2, 'z': 13},
                'two': {'c': 3},
                'three': [],
            },
            'non-root': [1, 2],
        }

        self.assertEqual(conf._data, result)

    def test_update_with_conf(self):
        """
        The update has to be capable to make recursive update.
        """
        conf = Conf(self.source_conf)

        conf.update({
            'root': Conf({
                'one': {'a': 42, 'z': 13},
                'three': [],
            }),
            'non-root': [1, 2],
        })

        result = {
            'root': Conf({
                'one': {'a': 42, 'b': 2, 'z': 13},
                'two': {'c': 3},
                'three': [],
            }),
            'non-root': [1, 2],
        }

        self.assertEqual(conf._data, result)

    def test_getitem_simple_key(self):
        """
        The __getitem__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        self.assertEqual(
            conf['root'], self.source_conf['root'])
        self.assertEqual(
            conf['root']['one'], self.source_conf['root']['one'])
        self.assertEqual(
            conf['root']['one']['a'], self.source_conf['root']['one']['a'])
        self.assertEqual(
            conf['root']['one']['b'], self.source_conf['root']['one']['b'])
        self.assertEqual(
            conf['root']['two'], self.source_conf['root']['two'])
        self.assertEqual(
            conf['root']['two']['c'], self.source_conf['root']['two']['c'])

        self.assertIsInstance(conf['root'], Conf)
        self.assertIsInstance(conf['root']['one'], Conf)
        self.assertIsInstance(conf['root']['one']['a'], int)
        self.assertIsInstance(conf['root']['one']['b'], int)
        self.assertIsInstance(conf['root']['two'], Conf)
        self.assertIsInstance(conf['root']['two']['c'], int)

        self.assertRaises(KeyError, lambda: conf['non-root'])

    def test_getitem_compound_key(self):
        """
        The __getitem__ has to handle compund keys correctly.
        """
        conf = Conf(self.source_conf)

        self.assertEqual(
            conf['root.one'], self.source_conf['root']['one'])
        self.assertEqual(
            conf['root.one.a'], self.source_conf['root']['one']['a'])
        self.assertEqual(
            conf['root.one.b'], self.source_conf['root']['one']['b'])
        self.assertEqual(
            conf['root.two'], self.source_conf['root']['two'])
        self.assertEqual(
            conf['root.two.c'], self.source_conf['root']['two']['c'])

        self.assertIsInstance(conf['root.one'], Conf)
        self.assertIsInstance(conf['root.one.a'], int)
        self.assertIsInstance(conf['root.one.b'], int)
        self.assertIsInstance(conf['root.two'], Conf)
        self.assertIsInstance(conf['root.two.c'], int)

        self.assertRaises(KeyError, lambda: conf['root.three'])

    def test_get_simple_key(self):
        """
        The get method has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        self.assertEqual(
            conf.get('root'), self.source_conf['root'])
        self.assertEqual(
            conf.get('root').get('one'), self.source_conf['root']['one'])
        self.assertEqual(
            conf.get('non-root', 'default'), 'default')

        try:
            conf.get('non-root')
        except KeyError:
            self.fail('The `get` method MUST BE safe!')

    def test_get_compound_key(self):
        """
        The __getitem__ has to handle compund keys correctly.
        """
        conf = Conf(self.source_conf)

        self.assertEqual(
            conf.get('root.one'), self.source_conf['root']['one'])
        self.assertEqual(
            conf.get('root.one.a'), self.source_conf['root']['one']['a'])
        self.assertEqual(
            conf.get('root.one.b'), self.source_conf['root']['one']['b'])
        self.assertEqual(
            conf.get('root.two'), self.source_conf['root']['two'])
        self.assertEqual(
            conf.get('root.two.c'), self.source_conf['root']['two']['c'])
        self.assertEqual(
            conf.get('root.three', 'default'), 'default')

        try:
            conf.get('root.three')
        except KeyError:
            self.fail('The `get` method MUST BE safe!')

    def test_setitem_simple_key(self):
        """
        The __setitem__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)
        conf['non-root'] = 42
        conf['root']['three'] = 13

        self.assertEqual(conf._data['non-root'], 42)
        self.assertEqual(conf._data['root']['three'], 13)

    def test_setitem_compound_key(self):
        """
        The __setitem__ has to handle compund keys correctly.
        """
        conf = Conf(self.source_conf)

        try:
            conf['root.three'] = 42
            conf['root.four.d'] = 13
        except:
            self.fail('The __setitem__ MUST NOT fail!')

        self.assertEqual(conf._data['root']['three'], 42)
        self.assertEqual(conf._data['root']['four']['d'], 13)

    def test_delitem_simple_key(self):
        """
        The __detitem__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        del conf['root']['two']
        self.assertIn('root', conf)
        self.assertNotIn('two', conf['root'])

        del conf['root']
        self.assertNotIn('root', conf)

        def delconf(conf, key):
            del conf[key]
        self.assertRaises(KeyError, delconf, conf, 'non-root')

    def test_delitem_compound_key(self):
        """
        The __detitem__ has to handle compund keys correctly.
        """
        conf = Conf(self.source_conf)
        del conf['root.two']

        self.assertIn('root', conf)
        self.assertNotIn('root.two', conf)

        def delconf(conf, key):
            del conf[key]
        self.assertRaises(KeyError, delconf, conf, 'root.three')

    def test_contains_simple_key(self):
        """
        The __contains__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        self.assertIn('root', conf)
        self.assertIn('one', conf['root'])
        self.assertIn('two', conf['root'])

        self.assertNotIn('one', conf)
        self.assertNotIn('two', conf)

    def test_contains_compound_key(self):
        """
        The __contains__ has to handle compound keys correctly.
        """
        conf = Conf(self.source_conf)

        self.assertIn('root.one', conf)
        self.assertIn('root.two', conf)
        self.assertIn('root.one.a', conf)
        self.assertIn('root.one.b', conf)
        self.assertIn('root.two.c', conf)

        self.assertNotIn('root.one.c', conf)
        self.assertNotIn('root.three', conf)

    def test_iter(self):
        """
        The __iter__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        self.assertSequenceEqual(conf, self.source_conf)

    def test_len(self):
        """
        The __len__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        self.assertEqual(len(conf), len(self.source_conf))

    def test_str(self):
        """
        The __str__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        self.assertEqual(str(conf), str(conf._data))

    def test_format(self):
        """
        The __format__ has to behave exactly like a dict's one.
        """
        conf = Conf(self.source_conf)

        self.assertEqual("{0}".format(conf), str(conf))

    def test_eq_operator(self):
        """
        The __eq__ has to compare an instance with Confs and dicts.
        """
        conf_a = Conf(self.source_conf)
        conf_b = Conf(self.source_conf)

        self.assertEqual(conf_a, conf_b)
        self.assertEqual(conf_a, conf_b._data)

    def test_ne_operator(self):
        """
        The __ne__ has to compare an instance with Confs and dicts.
        """
        conf_a = Conf(self.source_conf)
        conf_b = Conf(self.source_conf)

        conf_b['non-root'] = 42
        self.assertNotEqual(conf_a, conf_b)
        self.assertNotEqual(conf_a, conf_b._data)



# coding: utf-8
"""
    dooku.decorator
    ~~~~~~~~~~~~~~~

    The module implements various useful decorators.

    :copyright: (c) 2014, Igor Kalnitsky
    :license: BSD, see LICENSE for details
"""


class cached_property(object):
    """
    Decorator that converts a method into a lazy property.

    The wrapped method is called the first time to retrieve the result and
    then the calculated result is used next time you access the value::

        class Holocron(object):

            @cached_property
            def jinja_env(self):
                # (create and configure jinja environment)
                return jinja_env


    .. admonition:: Implementation details

        The property is implemented as non-data descriptor. That mean, the
        descriptor is invoked if there's no entry with the same name in the
        instance's ``__dict__``.

        This trick helps us to get rid of the function call overhead.

    :param func:
        A method to be wrapped.
    """
    def __init__(self, func):
        self.func = func

        self.__name__ = func.__name__
        self.__module__ = func.__module__
        self.__doc__ = func.__doc__

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        if self.__name__ not in obj.__dict__:
            obj.__dict__[self.__name__] = self.func(obj)
        return obj.__dict__[self.__name__]
